#!/usr/bin/env python
#-*- coding:utf-8 -*-

from __future__ import print_function, unicode_literals, division, absolute_import

import getopt
import io
import sys
import os
import glob
try:
    from pynlpl.formats import folia
except:
    print("ERROR: pynlpl not found, please obtain PyNLPL from the Python Package Manager ($ sudo pip install pynlpl) or directly from github: $ git clone git://github.com/proycon/pynlpl.git", file=sys.stderr)
    sys.exit(2)

def usage():
    print("folia2annotatedtxt", file=sys.stderr)
    print("  by Maarten van Gompel (proycon)", file=sys.stderr)
    print("  Tilburg University / Radboud University Nijmegen", file=sys.stderr)
    print("  2012 - Licensed under GPLv3", file=sys.stderr)
    print("", file=sys.stderr)
    print("This conversion script reads a FoLiA XML document and produces a", file=sys.stderr)
    print("simple inline output format in which tokens are space-separated and annotations are separated by a pipe symbol (|)", file=sys.stderr)
    print("Note that only simple token annotations are supported and a lot", file=sys.stderr)
    print("of FoLiA data can not be intuitively expressed in a simple columned format!", file=sys.stderr)
    print("", file=sys.stderr)
    print("Usage: folia2annotatedtxt [options] -C [columns] file-or-dir1 file-or-dir2 ..etc..", file=sys.stderr)

    print("Parameters:", file=sys.stderr)
    print("  -c [columns]                 Comma separated list of desired column layout (mandatory), choose from:", file=sys.stderr)
    print("                               id      - output word ID", file=sys.stderr)
    print("                               text    - output the text of the word (the word itself)", file=sys.stderr)
    print("                               pos     - output PoS annotation class", file=sys.stderr)
    print("                               poshead - output PoS annotation head feature", file=sys.stderr)
    print("                               lemma   - output lemma annotation class", file=sys.stderr)
    print("                               sense   - output sense annotation class", file=sys.stderr)
    print("                               phon    - output phonetic annotation class", file=sys.stderr)
    print("                               senid   - output sentence ID", file=sys.stderr)
    print("                               parid   - output paragraph ID", file=sys.stderr)
    print("                               N     - word/token number (absolute)", file=sys.stderr)
    print("                               n     - word/token number (relative to sentence)", file=sys.stderr)
    print("Options:", file=sys.stderr)
    print("  -o [filename]                Output to a single output file instead of stdout", file=sys.stderr)
    print("  -O                           Output each file to similarly named file (.columns or .csv)", file=sys.stderr)
    print("  -e [encoding]                Output encoding (default: utf-8)", file=sys.stderr)
    print("  -S                           Output one sentence per line", file=sys.stderr)
    print("Parameters for processing directories:", file=sys.stderr)
    print("  -r                           Process recursively", file=sys.stderr)
    print("  -E [extension]               Set extension (default: xml)", file=sys.stderr)
    print("  -O                           Output each file to similarly named .txt file", file=sys.stderr)
    print("  -P                           Like -O, but outputs to current working directory", file=sys.stderr)
    print("  -q                           Ignore errors", file=sys.stderr)

class settings:
    output_header = True
    outputfile = None
    ignoreerrors = False
    autooutput = False
    extension = 'xml'
    recurse = False
    encoding = 'utf-8'
    sentenceperline = False
    columnconf = []

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:OPhHSc:x:E:rq", ["help", "csv"])
    except getopt.GetoptError as err:
        print(str(err), file=sys.stderr)
        usage()
        sys.exit(2)

    outputfile = None

    for o, a in opts:
        if o == '-c':
            for a in a.split(','):
                settings.columnconf.append(a)
        elif o == '-h':
            usage()
            sys.exit(0)
        elif o == '-H':
            settings.output_header = False
        elif o == '-e':
            settings.encoding = a
        elif o == '-o':
            outputfile = a
        elif o == '-O':
            settings.autooutput = True
        elif o == '-P':
            settings.autooutput = True
            settings.autooutput_cwd = True
        elif o == '-E':
            settings.extension = a
        elif o == '-r':
            settings.recurse = True
        elif o == '-q':
            settings.ignoreerrors = True
        elif o == '-S':
            settings.sentenceperline = True
        else:
            raise Exception("No such option: " + o)

    if not settings.columnconf:
        print("ERROR: No column configuration specified (use -c)", file=sys.stderr)
        usage()
        sys.exit(2)


    if args:
        if outputfile: outputfile = io.open(outputfile,'w',encoding=settings.encoding)
        for x in args:
            if os.path.isdir(x):
                processdir(x,outputfile)
            elif os.path.isfile(x):
                process(x, outputfile)
            else:
                print("ERROR: File or directory not found: " + x, file=sys.stderr)
                sys.exit(3)
        if outputfile: outputfile.close()
    else:
        print("ERROR: Nothing to do, specify one or more files or directories", file=sys.stderr)



def resize(s, i, spacing):
    if len(s) >= spacing[i]:
        s = s[0:spacing[i] - 1] + ' '
    elif len(s) < spacing[i]:
        s = s + (' ' * (spacing[i] - len(s)))
    #print '[' + s + ']', len(s), spacing[i]
    return s

def processdir(d, outputfile = None):
    print("Searching in  " + d, file=sys.stderr)
    for f in glob.glob(os.path.join(d ,'*')):
        if f[-len(settings.extension) - 1:] == '.' + settings.extension:
            process(f, outputfile)
        elif settings.recurse and os.path.isdir(f):
            processdir(f, outputfile)

def process(filename, outputfile=None):
    try:
        print("Processing " + filename, file=sys.stderr)
        doc = folia.Document(file=filename)
        prevsen = None
        prevpar = None

        if settings.autooutput:
            ext = '.txt'
            if filename[-len(settings.extension) - 1:].lower() == '.' +settings.extension:
                outfilename = filename[:-len(settings.extension) - 1] + ext
            else:
                outfilename += ext
            if settings.autooutput_cwd:
                outfilename = os.path.basename(outfilename)

            print(" Saving as " + outfilename, file=sys.stderr)
            outputfile = io.open(outfilename,'w',encoding=settings.encoding)


        wordnum = 0


        for i, w in enumerate(doc.words()):
            if settings.sentenceperline:
                if w.sentence() != prevsen and i > 0:
                    if outputfile:
                        outputfile.write("\n")
                    else:
                        print()
                wordnum = 0
            if w.paragraph() != prevpar and i > 0:
                if outputfile:
                    outputfile.write("\n")
                else:
                    print()
                wordnum = 0
            prevpar = w.paragraph()
            prevsen = w.sentence()
            wordnum += 1
            columns = []
            for c in settings.columnconf:
                if c == 'id':
                    columns.append(w.id)
                elif c == 'text':
                    columns.append(w.text())
                elif c == 'n':
                    columns.append(str(wordnum))
                elif c == 'N':
                    columns.append(str(i+1))
                elif c == 'pos':
                    try:
                        columns.append(w.annotation(folia.PosAnnotation).cls)
                    except:
                        columns.append('-')
                elif c == 'poshead':
                    try:
                        columns.append(w.annotation(folia.PosAnnotation).feat('head'))
                    except:
                        columns.append('-')
                elif c == 'lemma':
                    try:
                        columns.append(w.annotation(folia.LemmaAnnotation).cls)
                    except:
                        columns.append('-')
                elif c == 'sense':
                    try:
                        columns.append(w.annotation(folia.SenseAnnotation).cls)
                    except:
                        columns.append('-')
                elif c == 'phon':
                    try:
                        columns.append(w.annotation(folia.PhonAnnotation).cls)
                    except:
                        columns.append('-')
                elif c == 'senid':
                    columns.append(w.sentence().id)
                elif c == 'parid':
                    try:
                        columns.append(w.paragraph().id)
                    except:
                        columns.append('-')
                elif c:
                    print("ERROR: Unsupported configuration: " + c, file=sys.stderr)
                    sys.exit(1)


            word = "|".join(columns).strip()
            if outputfile:
                if wordnum > 1: outputfile.write(" ")
                outputfile.write(word)
            else:
                if wordnum > 1: print(" ", end="")
                if sys.version < '3':
                    print(word.encode(settings.encoding),end="")
                else:
                    print(word,end="")

        if settings.autooutput:
            outputfile.close()
        elif outputfile:
            outputfile.flush()
    except Exception as e:
        if settings.ignoreerrors:
            print("ERROR: An exception was raised whilst processing " + filename, e, file=sys.stderr)
        else:
            raise

if __name__ == "__main__":
    main()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

from __future__ import print_function, unicode_literals, division, absolute_import

import getopt
import sys
import os
import glob
try:
    from pynlpl.formats import folia
except:
    print("ERROR: pynlpl not found, please obtain PyNLPL from the Python Package Manager ($ sudo easy_install pynlpl) or directly from github: $ git clone git://github.com/proycon/pynlpl.git",file=sys.stderr)
    sys.exit(2)

def usage():
    print("foliaquery",file=sys.stderr)
    print("  by Maarten van Gompel (proycon)",file=sys.stderr)
    print("  Tilburg University / Radboud University Nijmegen",file=sys.stderr)
    print("  2012 - Licensed under GPLv3",file=sys.stderr)
    print("",file=sys.stderr)
    print("Query one or more FoLiA documents for certain patterns.",file=sys.stderr)
    print("",file=sys.stderr)
    print("Usage: foliaquery [options] file-or-dir1 file-or-dir2 ..etc..",file=sys.stderr)
    print("",file=sys.stderr)
    print("Parameters for output:"        ,file=sys.stderr)
    print("  --text \"[words]\"           Text pattern (Case sensitive)",file=sys.stderr)
    print("  --pos \"[postags]\"          Pos pattern",file=sys.stderr)
    print("  --lemma \"[lemmas]\"         Lemma pattern",file=sys.stderr)
    print("  --sense \"[sense]\"          Sense pattern",file=sys.stderr)
    print("  --phon \"[phon]\"            Phonetic pattern",file=sys.stderr)
    print("  -i                           Patterns are case-insensitive",file=sys.stderr)
    print("  -L [length]                  Left-context size words/tokens (default: 0)",file=sys.stderr)
    print("  -R [length]                  Right-context size in words/tokens (default: 0)"    ,file=sys.stderr)
    print("  -e [encoding]                Output encoding (default: utf-8)"     ,file=sys.stderr)
    print("Parameters for processing directories:",file=sys.stderr)
    print("  -r                           Process recursively",file=sys.stderr)
    print("  -E [extension]               Set extension (default: xml)",file=sys.stderr)
    print("  -q                           Ignore errors",file=sys.stderr)
    print("",file=sys.stderr)
    print("Pattern syntax:",file=sys.stderr)
    print("    Fixed-width wildcard: ^ ",file=sys.stderr)
    print("    Variable-width wildcard: * ",file=sys.stderr)
    print("    Disjunction: | ",file=sys.stderr)
    print("    Regular Expression: {REGEXP}",file=sys.stderr)

    print(""    ,file=sys.stderr)
    print("Examples:",file=sys.stderr)
    print("   1) foliaquery --text=\"to be * to be\"",file=sys.stderr)
    print("       Matches any gap of any size (up to the maximum)"    ,file=sys.stderr)
    print("   2) foliaquery --text=\"to be ^ ^ to be\""   ,file=sys.stderr)
    print("       Matches any gap of exactly two tokens",file=sys.stderr)
    print("   3) foliaquery --pos=\"ADJ NOUN\"",file=sys.stderr)
    print("       Searching by annotation"    ,file=sys.stderr)
    print("   4) foliaquery --text=\"rent\" --pos=\"NOUN\"",file=sys.stderr)
    print("       Patterns may be combined, matches have to satisfy all patterns",file=sys.stderr)
    print("   5) foliaquery --text=\"he leaves|departs today|tomorrow\" --pos=\"PRON VERB ^\"",file=sys.stderr)
    print("       The pipe character allows for disjunctions in single tokens",file=sys.stderr)
    print("   6a) foliaquery --text=\"we {w[io]n}\" --pos=\"PRON VERB\"",file=sys.stderr)
    print("   6b) foliaquery --text=\"{.*able}\" --pos=\"ADJ\"",file=sys.stderr)
    print("       Curly braces specify a regular expression for a single token"    ,file=sys.stderr)




def parsepattern(rawpattern, annotationtype): #, annotationset=None):
    components = []
    for tokenpattern in rawpattern.strip().split(' '):
        if tokenpattern == '*':
            components.append('*')
        elif tokenpattern == '^':
            components.append(True)
        elif tokenpattern[0] == '{' and tokenpattern[-1] == '}':
            components.append( folia.RegExp(tokenpattern[1:-1]) )
        elif '|' in tokenpattern:
            components.append( tuple(tokenpattern.split('|')) )
        else:
            components.append(tokenpattern)
    d = {'casesensitive':settings.casesensitive}
    if annotationtype:
        d['matchannotation'] = annotationtype
    return folia.Pattern(*components,**d) #, matchannotationset=annotationset)



def process(filename, patterns):
    try:
        print("Processing " + filename, file=sys.stderr)
        doc = folia.Document(file=filename)
        for match in doc.findwords(*patterns ):
            s = ""
            for token in match:
                s += "\t" + token.text()
            s = filename + "\t" + match[0].id + s
            if sys.version < '3':
                print(s.encode(settings.encoding))
            else:
                print(s)
    except Exception as e:
        if settings.ignoreerrors:
            print("ERROR: An exception was raised whilst processing " + filename + ":", e            ,file=sys.stderr)
        else:
            raise


def processdir(d, patterns):
    print("Searching in  " + d,file=sys.stderr)
    for f in glob.glob(os.path.join(d,'*')):
        if f[-len(settings.extension) - 1:] == '.' + settings.extension:
            process(f, patterns)
        elif settings.recurse and os.path.isdir(f):
            processdir(f, patterns)


class settings:
    leftcontext = 0
    rightcontext = 0

    extension = 'xml'
    recurse = False
    encoding = 'utf-8'

    ignoreerrors = False
    casesensitive = True


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:OE:hqr", ["help","text=","pos=","lemma=","sense=","phon="])
    except getopt.GetoptError as err:
        print(str(err), file=sys.stderr)
        usage()
        sys.exit(2)



    patterns = []

    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '--text':
            patterns.append( parsepattern(a, None) )
        elif o == '--pos':
            patterns.append( parsepattern(a, folia.PosAnnotation) )
        elif o == '--lemma':
            patterns.append( parsepattern(a, folia.LemmaAnnotation) )
        elif o == '--sense':
            patterns.append( parsepattern(a, folia.SenseAnnotation) )
        elif o == '--phon':
            patterns.append( parsepattern(a, folia.PhonAnnotation) )


        elif o == '-e':
            settings.encoding = a
        elif o == '-E':
            settings.extension = a
        elif o == '-r':
            settings.recurse = True
        elif o == '-q':
            settings.ignoreerrors = True
        else:
            raise Exception("No such option: " + o)


    if args:
        for x in args:
            if os.path.isdir(x):
                processdir(x, patterns)
            elif os.path.isfile(x):
                process(x, patterns)
            elif x[0:2] != '--':
                print("ERROR: File or directory not found: " + x,file=sys.stderr)
                sys.exit(3)
    else:
        print("ERROR: Nothing to do, specify one or more files or directories",file=sys.stderr)

if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf8 -*-

from __future__ import print_function, unicode_literals, division, absolute_import
import foliatools.xslt as xslt

def main():
    usage = """folia2dcoi
  by Maarten van Gompel (proycon)
  Tilburg University / Radboud University Nijmegen
  2012 - Licensed under GPLv3

This conversion script converts one or more FoLiA documents to D-Coi XML format, omitting
any annotations that can not be represented in the D-Coi format.

Usage: folia2dcoi [options] file-or-dir1 file-or-dir2 ..etc.."""

    xslt.main('folia2dcoi.xsl','dcoi.xml', usage)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-


from __future__ import print_function, unicode_literals, division, absolute_import

import getopt
import sys
try:
    from pynlpl.formats import folia
except:
    print("ERROR: pynlpl not found, please obtain PyNLPL from the Python Package Manager ($ sudo easy_install pynlpl) or directly from github: $ git clone git://github.com/proycon/pynlpl.git",file=sys.stderr)
    sys.exit(2)

def usage():
    print("foliamerge",file=sys.stderr)
    print("  by Maarten van Gompel (proycon)",file=sys.stderr)
    print("  Radboud University Nijmegen",file=sys.stderr)
    print("  2016 - Licensed under GPLv3",file=sys.stderr)
    print("",file=sys.stderr)
    print("Merges annotations from two or more FoLiA documents. Structural elements are never added. Annotations can only be merged if their parent elements have IDs.",file=sys.stderr)
    print("",file=sys.stderr)
    print("Usage: foliamerge [options] file1 file2 file3 ... ",file=sys.stderr)
    print("",file=sys.stderr)
    print("Options:",file=sys.stderr)
    print("  -o [file]                    Output file",file=sys.stderr)
    print("  -s                           Substitute: use first input file as output as well",file=sys.stderr)


def mergechildren(parent, outputdoc):
    merges = 0
    for e in parent:
        if (isinstance(e, folia.AbstractAnnotation) or isinstance(e, folia.AbstractAnnotationLayer)) and parent.id:
            try:
                e.ANNOTATIONTYPE
            except:
                continue

            if (e.ANNOTATIONTYPE, e.set) in outputdoc.annotations:
                assert e.parent == parent
                try:
                    newparent = outputdoc[parent.id]
                except:
                    pass
                #check if the annotation already exists
                if isinstance(e, folia.AbstractTokenAnnotation) and newparent.hasannotation(e.__class__, e.set):
                    print("Annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id + " , already exists... skipping", file=sys.stderr)
                    pass
                elif isinstance(e, folia.AbstractAnnotationLayer) and newparent.hasannotationlayer(e.__class__, e.set):
                    print("Annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id + " , already exists... skipping", file=sys.stderr)
                    pass
                else:
                    print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id, file=sys.stderr)
                    c = e.copy(outputdoc) #make a copy, linked to outputdoc
                    newparent.append(c) #append to outputdoc
                    merges += 1
        elif isinstance(e, folia.AbstractElement):
            merges += mergechildren(e, outputdoc)
    return merges



def foliamerge(outputfile, *files):
        outputdoc = None
        merges = 0

        for i, filename in enumerate(files):
            print("Processing " + filename, file=sys.stderr)
            inputdoc = folia.Document(file=filename)
            if i == 0:
                 print("(pivot document)",file=sys.stderr)
                 outputdoc = inputdoc
            else:
                print("(merging document)",file=sys.stderr)

                for annotationtype,set in inputdoc.annotations:
                    if not outputdoc.declared(annotationtype,set):
                        outputdoc.declare( annotationtype, set)

                for e in inputdoc:
                    merges += mergechildren(e, outputdoc)

        if outputfile and merges > 0:
            outputdoc.save(outputfile)

        return outputdoc

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:sh", ["help"])
    except getopt.GetoptError as err:
        print(str(err),file=sys.stderr)
        usage()
        sys.exit(2)

    outputfile = None
    substitute = False

    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '-o':
            outputfile = a
        elif o == '-s':
            substitute = True
        else:
            raise Exception("No such option: " + o)

    if len(args) < 2:
        print("ERROR: At least two files need to be specified",file=sys.stderr)
        sys.exit(2)

    if substitute:
        outputfile = args[0]

    outputdoc = foliamerge(outputfile, *args)
    if not outputfile:
        xml = outputdoc.xmlstring()
        if sys.version < '3':
            if isinstance(xml,unicode):
                print(xml.encode('utf-8'))
            else:
                print(xml)
        else:
            print(xml)


if __name__ == "__main__":
    main()






#!/usr/bin/env python
# -*- coding: utf8 -*-

from __future__ import print_function, unicode_literals, division, absolute_import

import getopt
import sys
import os
import glob
import traceback
try:
    from pynlpl.formats import folia
except:
    print("ERROR: pynlpl not found, please obtain PyNLPL from the Python Package Manager ($ sudo easy_install pynlpl) or directly from github: $ git clone git://github.com/proycon/pynlpl.git", file=sys.stderr)
    sys.exit(2)

def usage():
    print("foliacorrect", file=sys.stderr)
    print("  by Maarten van Gompel (proycon)", file=sys.stderr)
    print("  Radboud University Nijmegen", file=sys.stderr)
    print("  2015 - Licensed under GPLv3", file=sys.stderr)
    print("", file=sys.stderr)
    print("FoLiA " + folia.FOLIAVERSION + ", library version " + folia.LIBVERSION, file=sys.stderr)
    print("", file=sys.stderr)
    print("Corrections are one of the most complex forms of annotation in FoLiA. It may occur that you want to strip explicit corrections from a document, and leave only either the corrected or original version. The document can then be parsed by simpler parsers that can not handle explicit corrections. Alternatively, you may want to accept the best suggestion for correction (with or without stripping the correction element). This tool provides the means to do all that.", file=sys.stderr)
    print("", file=sys.stderr)
    print("Usage: foliacorrect [options] file-or-dir1 file-or-dir2 ..etc..", file=sys.stderr)
    print("", file=sys.stderr)
    print("Parameters for processing directories:", file=sys.stderr)
    print("  -r                           Process recursively", file=sys.stderr)
    print("  -E [extension]               Set extension (default: xml)", file=sys.stderr)
    print("  -V                           Show version info", file=sys.stderr)
    print("  -q                           Ignore errors",file=sys.stderr)
    print("  --corrected                  Keep the corrected versions, removing all explicit corrections", file=sys.stderr)
    print("  --original                   Keep the original versions,  removing all explicit corrections", file=sys.stderr)
    print("  --acceptsuggestion           Automatically accept the suggestion with the hightest confidence (can not be used with --original, can be used with --corrected)", file=sys.stderr)
    print("  --set                        Correction set to filter on", file=sys.stderr)
    print("  --class                      Correction class to filter on", file=sys.stderr)
    print("  --print                      Print all corrections, do not change the document", file=sys.stderr)



def replace(correction, correctionchild):
    parent = correction.parent
    index = parent.getindex(correction)
    elements = correctionchild.copychildren(correction.doc)
    parent.remove(correction)
    for i, e in enumerate(elements):
        if isinstance(e, folia.TextContent) and e.cls == 'original':
            e.cls = 'current'
        parent.insert(index+i, e)



def correct(filename,corrected, original, acceptsuggestion, setfilter,classfilter, output):
    changed = False
    try:
        doc = folia.Document(file=filename)
        for text in doc:
            for correction in list(text.select(folia.Correction, setfilter)):
                if not classfilter or correction.cls == classfilter:
                    if original:
                        if correction.hasoriginal():
                            #restore original
                            print("Restoring original version for " + str(correction.id),file=sys.stderr)
                            replace(correction, correction.original())
                            changed = True
                        elif correction.hasoriginal(True): #check for empty original
                            #insertion, remove it
                            correction.parent.remove(correction)
                    elif corrected:
                        if correction.hasnew():
                            print("Keeping corrected version for " + str(correction.id),file=sys.stderr)
                            replace(correction, correction.new())
                            changed = True
                    elif correction.hassuggestions() and acceptsuggestion:
                        bestsuggestion = None
                        changed = True
                        for suggestion in correction.hassuggestions():
                            if not bestsuggestion or (suggestion.confidence and not bestsuggestion.confidence) or (suggestion.confidence and bestsuggestion.confidence and suggestion.confidence > bestsuggestion.confidence):
                                bestsuggestion = suggestion
                        if bestsuggestion:
                            if corrected:
                                replace(correction, bestsuggestion)
                            else:
                                raise NotImplementedError #TODO
                    if output:
                        print(correction.xmlstring())
        if changed:
            if settings.stdout:
                print(doc.xmlstring())
            else:
                doc.save()
    except Exception as e:
        if settings.ignoreerrors:
            print("ERROR: An exception was raised whilst processing " + filename + ":", e, file=sys.stderr)
        else:
            raise




def processdir(d,corrected, original, acceptsuggestion, setfilter,classfilter,output):
    print("Searching in  " + d,file=sys.stderr)
    for f in glob.glob(os.path.join(d ,'*')):
        if f[-len(settings.extension) - 1:] == '.' + settings.extension:
            correct(f,corrected,original,  acceptsuggestion, setfilter,classfilter,output)
        elif settings.recurse and os.path.isdir(f):
            processdir(f, corrected,original, acceptsuggestion, setfilter,classfilter,output)


class settings:
    extension = 'xml'
    recurse = False
    encoding = 'utf-8'
    ignoreerrors = False
    stdout = False

def main():
    original = acceptsuggestion = output = corrected = False
    setfilter = classfilter = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], "E:srhqVO", ["help","original","corrected","acceptsuggestion","set=","class=","print"])
    except getopt.GetoptError as err:
        print(str(err), file=sys.stderr)
        usage()
        sys.exit(2)

    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '-E':
            settings.extension = a
        elif o == '-r':
            settings.recurse = True
        elif o == '-q':
            settings.ignoreerrors = True
        elif o == '-O':
            settings.stdout = True
        elif o == '-V':
            print("FoLiA " + folia.FOLIAVERSION + ", library version " + folia.LIBVERSION,file=sys.stderr)
            sys.exit(0)
        elif o == '--original':
            original = True
        elif o == '--corrected':
            corrected = True
        elif o == '--acceptsuggestion':
            acceptsuggestion = True
        elif o == '--set' or o == '--set=':
            setfilter = a
        elif o == '--class' or o == '--class=':
            classfilter = a
        elif o == '--print':
            output = True
        else:
            raise Exception("No such option: " + o)


    if len(args) >= 1:
        for x in sys.argv[1:]:
            if x[0] != '-':
                if os.path.isdir(x):
                    processdir(x,corrected,original, acceptsuggestion, setfilter,classfilter,output)
                elif os.path.isfile(x):
                    correct(x,corrected, original, acceptsuggestion, setfilter,classfilter,output)
                else:
                    print("ERROR: File or directory not found: " + x,file=sys.stderr)
                    sys.exit(3)
    else:
        print("ERROR: No files specified",file=sys.stderr)
        sys.exit(2)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#Generate library specification code (for either Python or C++) on the the basis of folia.yml
#Used by respectively pynlpl and libfolia

from __future__ import print_function, unicode_literals, division, absolute_import

import sys
import datetime
import os
from collections import defaultdict
import yaml



skip_properties = {
    'c++': ('textcontainer','phoncontainer','primaryelement'), #these are not handled in libfolia, or handled differently, don't output these in the source
}

#Load specification
specfiles= [  os.path.join(os.path.dirname(__file__) ,'../schemas/folia.yml'), 'folia.yml' ]
spec = None
for specfile in specfiles:
    spec = yaml.load(open(specfile,'r'))
    break

if spec is None:
    print("FoLiA Specification file folia.yml could not be found in " + ", ".join(specfiles) ,file=sys.stderr)


parents = defaultdict(list)

elementdict = {} #flat (unnested) dictionary

def getelements(d):
    elements = []
    if 'elements' in d:
        for e in d['elements']:
            elementdict[e['class']] = e
            elements.append(e)
            children = getelements(e)
            elements += children
            for c in children:
                if e['class'] not in parents[c['class']]:
                    parents[c['class']].append(e['class'])
    return elements

elements = getelements(spec) #gathers all class names
elements.sort(key=lambda x: x['class'])
elementnames = [ e['class'] for e in elements ]


################################################################

def addfromparents(elementname, key):
    value = set(spec['defaultproperties']['accepted_data'])
    if 'properties' in elementdict[elementname] and key in elementdict[elementname]['properties'] and elementdict[elementname]['properties'][key]:
        value |= set(elementdict[elementname]['properties'][key])
    else:
        value |= set()
    for parent in parents[elementname]:
        value |= addfromparents(parent, key)
    return value



def outputvar(var, value, target, declare = False):
    """Output a variable ``var`` with value ``value`` in the specified target language."""

    #do we need to quote the value? (bool)
    varname = var.split('.')[-1]

    if isinstance(value, str) and varname.upper() in ('ACCEPTED_DATA','REQUIRED_DATA','REQUIRED_ATTRIBS', 'OPTIONAL_ATTRIBS','ANNOTATIONTYPE'):
        quote = False
    else:
        quote = True


    if isinstance(value, str):
        value = value.replace("\n","\\n").replace("\t","\\t")

    if target == 'python':
        if varname == 'ANNOTATIONTYPE' and isinstance(value,str):
            value = 'AnnotationType.' + value

        if value is None:
                return var + ' = None'
        elif isinstance(value, bool):
            if value:
                return var + ' = True'
            else:
                return var + ' = False'
        elif isinstance(value, (int, float) ):
            return var + ' = ' + str(value)
        elif isinstance(value, (list,tuple,set) ):
            if varname in ('ACCEPTED_DATA','REQUIRED_DATA') or  all([ x in elementnames for x in value ]):
                return var + ' = (' + ', '.join(value) + ',)'
            elif all([ x in spec['attributes'] for x in value ]):
                return var + ' = (' + ', '.join(['Attrib.' + x for x in value]) + ',)'

            if len(value) == 0:
                return var + ' = ()'

            #list items are  enums or classes, never string literals
            if quote:
                return var + ' = (' + ', '.join([ '"' + x + '"' for x in value]) + ',)'
            else:
                return var + ' = (' + ', '.join(value) + ',)'
        else:
            if quote:
                return var + ' = "' + value  + '"'
            else:
                return var + ' = ' + value
    elif target == 'c++':
        typedeclaration = ''
        if value is None:
            if declare: raise NotImplementedError("Declare not supported for None values")
            if varname in ('REQUIRED_ATTRIBS','OPTIONAL_ATTRIBS'):
                return var + ' = NO_ATT;'
            elif varname == 'ANNOTATIONTYPE':
                return var + ' = AnnotationType::NO_ANN;'
            elif varname in ('XMLTAG','TEXTDELIMITER'):
                return var + ' = "NONE";'
            elif varname  == 'REQUIRED_DATA':
                return var + ' = {};'
            elif varname  == 'SUBSET':
                return var + ' = "";'
            else:
                raise NotImplementedError("Don't know how to handle None for " + var)
        elif isinstance(value, bool):
            if declare: typedeclaration = 'const bool '
            if value:
                return typedeclaration + var + ' = true;'
            else:
                return typedeclaration + var + ' = false;'
        elif isinstance(value, int ):
            if declare: typedeclaration = 'const int '
            return typedeclaration + var + ' = ' + str(value) + ';'
        elif isinstance(value, float ):
            if declare: typedeclaration = 'const double '
            return typedeclaration + var + ' = ' + str(value) + ';'
        elif isinstance(value, (list,tuple,set)):
            #list items are  enums or classes, never string literals
            if varname in ('ACCEPTED_DATA','REQUIRED_DATA') or  all([ x in elementnames for x in value ]):
                if declare:
                    typedeclarion = 'const set<ElementType> '
                    operator = '='
                else:
                    typedeclaration = ''
                    operator = '+='
                value = [ x + '_t' for x in value ]
                return typedeclaration + var + ' ' + operator + ' {' + ', '.join(value) + '};'
            elif all([ x in spec['attributes'] for x in value ]):
                return var + ' = ' + '|'.join(value) + ';'
            else:
                return typedeclaration + var + ' = { ' + ', '.join([ '"' + x + '"' for x in value if x]) + ', };'
        else:
            if varname == 'ANNOTATIONTYPE':
                value = "AnnotationType::" + value

            if quote:
                if declare: typedeclaration = 'const string '
                return typedeclaration + var + ' = "' + value+ '";'
            else:
                if declare: typedeclaration = 'const auto '
                return typedeclaration + var + ' = ' + value+ ';'

#concise description for all available template blocks
blockhelp = {
        'namespace': 'The FoLiA XML namespace',
        'version': 'The FoLiA version',
        'version_major': 'The FoLiA version (major)',
        'version_minor': 'The FoLiA version (minor)',
        'version_sub': 'The FoLiA version (sub/rev)',
        'attributes': 'Defines all common FoLiA attributes (as part of the Attrib enumeration)',
        'annotationtype': 'Defines all annotation types (as part of the AnnotationType enumeration)',
        'instantiateelementproperties': 'Instantiates all element properties for the first time, setting them to the default properties',
        'setelementproperties': 'Sets all element properties for all elements',
        'annotationtype_string_map': 'A mapping from annotation types to strings',
        'string_annotationtype_map': 'A mapping from strings to annotation types',
        'annotationtype_xml_map': 'A mapping from annotation types to xml tags (strings)',
        'structurescope': 'Structure scope above the sentence level, used by next() and previous() methods',
        'defaultproperties': 'Default properties which all elements inherit',
        'default_ignore': 'Default ignore list for the select() method, do not descend into these',
        'default_ignore_annotations': 'Default ignore list for token annotation',
        'default_ignore_structure': 'Default ignore list for structure annotation',
}

def setelementproperties_cpp(element,indent, defer,done):
    commentsign = "//"
    target = 'c++'
    s = commentsign + "------ " + element['class'] + " -------\n"
    if element['class'] in parents:
        for parent in parents[element['class']]:
            if parent not in done:
                defer[parent].append(element)
                return None
            else:
                s += indent + element['class'] + '::PROPS = ' + parent + '::PROPS;\n'
            break
    s += indent + element['class'] + '::PROPS.ELEMENT_ID = ' + element['class'] + '_t;\n'
    if 'properties' in element:
        for prop, value in sorted(element['properties'].items()):
            if target not in skip_properties or prop not in skip_properties[target]:
                if prop == 'xmltag':
                    if 'Feature' in parents[element['class']] and 'subset' in element['properties'] and element['properties']['subset']:
                        value = element['properties']['subset']
                elif prop == 'accepted_data':
                    value = tuple(sorted(addfromparents(element['class'],'accepted_data')))
                    if ('textcontainer' in element['properties'] and element['properties']['textcontainer']) or ('phoncontainer' in element['properties'] and element['properties']['phoncontainer']):
                        value += ('XmlText',)
                    if 'WordReference' in value:
                        value += ('Word','Morpheme','Phoneme')
                s += indent + outputvar(element['class'] + '::PROPS.' + prop.upper(),  value, target) + '\n'
    done[element['class']] = True
    return s

def flattenclasses(candidates):
    done = {}
    resolved = set()
    for c in candidates:
        for child, parentlist in parents.items():
            if c in parentlist:
                if child not in done and child not in candidates:
                    candidates.append(child)
        if c[:8] != 'Abstract':
            resolved.add(c)
    return resolved




def outputblock(block, target, varname, indent = ""):
    """Output the template block (identified by ``block``) for the target language"""

    if target == 'python':
        commentsign = '#'
    elif target == 'c++':
        commentsign = '//'
    else:
        raise NotImplementedError("Unknown target language: " + target)

    if block in blockhelp:
        s = indent + commentsign + blockhelp[block] + "\n" #output what each block does
    else:
        s = ''

    if block == 'header':
        s += indent + commentsign + "This file was last updated according to the FoLiA specification for version " + str(spec['version']) + " on " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + ", using foliaspec.py\n"
        s += indent + commentsign + "Code blocks after a foliaspec comment (until the next newline) are automatically generated. **DO NOT EDIT THOSE** and **DO NOT REMOVE ANY FOLIASPEC COMMENTS** !!!"
    elif block == 'version_major':
        versionfields = [ int(x) for x in spec['version'].split('.') ]
        s += indent + outputvar(varname, versionfields[0], target, True)
    elif block == 'version_minor':
        versionfields = [ int(x) for x in spec['version'].split('.') ]
        s += indent + outputvar(varname, versionfields[1] if len(versionfields) > 1 else 0, target, True)
    elif block == 'version_sub' or block == 'version_rev':
        versionfields = [ int(x) for x in spec['version'].split('.') ]
        s += indent + outputvar(varname, versionfields[2] if len(versionfields) > 2 else 0, target, True)
    elif block == 'attributes':
        if target == 'python':
            s += indent + "class Attrib:\n"
            s += indent + "    " +  ", ".join(spec['attributes']) + " = range(" + str(len(spec['attributes'])) + ")"
        elif target == 'c++':
            s += indent + "enum Attrib : int { NO_ATT=0, "
            value = 1
            for attrib in spec['attributes']:
                s +=  attrib + '=' + str(value) + ', '
                value *= 2
            s += 'ALL='+str(value) + ' };'
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'elementtype':
        if target == 'c++':
            s += indent + "enum ElementType : unsigned int { BASE=0,"
            s += ", ".join([ e + '_t' for e in elementnames]) + ", PlaceHolder_t, XmlComment_t, XmlText_t,  LastElement };\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'annotationtype':
        if target == 'python':
            s += indent + "class AnnotationType:\n"
            s += indent + "    " +  ", ".join(spec['annotationtype']) + " = range(" + str(len(spec['annotationtype'])) + ")"
        elif target == 'c++':
            s += indent + "enum AnnotationType : int { NO_ANN,"
            s += ", ".join(spec['annotationtype']) + ", LAST_ANN };\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'defaultproperties':
        if target == 'c++':
            s += indent + "ELEMENT_ID = BASE;\n"
            s += indent + "ACCEPTED_DATA.insert(XmlComment_t);\n"
            for prop, value in sorted(spec['defaultproperties'].items()):
                if target not in skip_properties or prop not in skip_properties[target]:
                    s += indent + outputvar( prop.upper(),  value, target) + '\n'
        elif target == 'python':
            for prop, value in sorted(spec['defaultproperties'].items()):
                s += indent + outputvar('AbstractElement.' + prop.upper(),  value, target) + '\n'
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'instantiateelementproperties':
        if target == 'c++':
            for element in elements:
                s += indent + "properties " + element['class'] + '::PROPS = DEFAULT_PROPERTIES;\n'
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'setelementproperties':
        if target == 'python':
            for element in elements:
                s += commentsign + "------ " + element['class'] + " -------\n"
                if 'properties' in element:
                    for prop, value in sorted(element['properties'].items()):
                        if prop == 'accepted_data':
                            value = tuple(sorted(addfromparents(element['class'],'accepted_data')))
                        s += indent + outputvar(element['class'] + '.' + prop.upper(),  value, target) + '\n'
        elif target == 'c++':
            done = {}
            defer = defaultdict(list) #defer output of some elements until parent elements are processed:  hook => deferred_elements
            for element in elements:
                output = setelementproperties_cpp(element,indent, defer,done)
                if output:
                    s += output
                    if element['class'] in defer:
                        for deferred in defer[element['class']]:
                            s += setelementproperties_cpp(deferred,indent, defer,done)

        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'annotationtype_string_map':
        if target == 'c++':
            s += indent + "const map<AnnotationType::AnnotationType,string> ant_s_map = {\n"
            s += indent + "  { AnnotationType::NO_ANN, \"NONE\" },\n"
            done = {}
            for element in elements:
                if 'properties' in element and  'annotationtype' in element['properties'] and element['properties']['annotationtype'] not in done:
                    #if 'primaryelement' in element['properties'] and not element['properties']['primaryelement']: continue #not primary, skip
                    s += indent + "  { AnnotationType::" + element['properties']['annotationtype'] + ',  "' + element['properties']['annotationtype'].lower() + '" },\n'
                    done[element['properties']['annotationtype']] = True #prevent duplicates
            s += indent + "};\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'string_annotationtype_map':
        if target == 'c++':
            s += indent + "const map<string,AnnotationType::AnnotationType> s_ant_map = {\n"
            s += indent + "  { \"NONE\", AnnotationType::NO_ANN },\n"
            done = {}
            for element in elements:
                if 'properties' in element and  'annotationtype' in element['properties'] and element['properties']['annotationtype'] not in done:
                    s += indent + '  { "' + element['properties']['annotationtype'].lower() + '", AnnotationType::' + element['properties']['annotationtype'] + ' },\n'
                    done[element['properties']['annotationtype']] = True #prevent duplicates
            s += indent + "};\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'annotationtype_xml_map':
        if target == 'python':
            s += indent + "ANNOTATIONTYPE2XML = {\n"
            for element in elements:
                if 'properties' in element and 'xmltag' in element['properties'] and element['properties']['xmltag'] and 'annotationtype' in element['properties']:
                    if 'primaryelement' in element['properties'] and not element['properties']['primaryelement']: continue #not primary, skip
                    s += indent + "    AnnotationType." + element['properties']['annotationtype'] + ':  "' + element['properties']['xmltag'] + '" ,\n'
            s += indent + "}"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'elementtype_string_map':
        if target == 'c++':
            s += indent + "const map<ElementType,string> et_s_map = {\n"
            s += indent + "  { BASE, \"FoLiA\" },\n"
            for element in elements:
                if 'properties' in element and 'xmltag' in element['properties'] and element['properties']['xmltag']:
                    s += indent + "  { " + element['class'] + '_t,  "' + element['properties']['xmltag'] + '" },\n'
                elif 'properties' in element and 'subset' in element['properties'] and element['properties']['subset']:
                    if element['class'] == 'HeadFeature':
                        s += indent + "  { HeadFeature_t,  \"headfeature\" },\n"
                    else:
                        s += indent + "  { " + element['class'] + '_t,  "' + element['properties']['subset'] + '" },\n'
                else:
                    s += indent + "  { " + element['class'] + '_t,  "_' + element['class'] + '" },\n'
            s += indent + '  { PlaceHolder_t, "_PlaceHolder" },\n'
            s += indent + '  { XmlComment_t, "_XmlComment" },\n'
            s += indent + '  { XmlText_t, "_XmlText" }\n'
            s += indent + "};\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'string_elementtype_map':
        if target == 'c++':
            s += indent + "const map<string,ElementType> s_et_map = {\n"
            s += indent + "  { \"FoLiA\", BASE },\n"
            for element in elements:
                if 'properties' in element and 'xmltag' in element['properties'] and element['properties']['xmltag']:
                    s += indent + '  { "' + element['properties']['xmltag'] + '", ' + element['class'] + '_t  },\n'
                elif 'properties' in element and 'subset' in element['properties'] and element['properties']['subset']:
                    if element['class'] == 'HeadFeature':
                        s += indent + "  { \"headfeature\", HeadFeature_t },\n"
                    else:
                        s += indent + '  { "' + element['properties']['subset'] + '", ' + element['class'] + '_t  },\n'
                else:
                    s += indent + '  { "_' + element['class'] + '", ' + element['class'] + '_t  },\n'
            s += indent + '  { "_PlaceHolder", PlaceHolder_t  },\n'
            s += indent + '  { "_XmlComment", XmlComment_t  },\n'
            s += indent + '  { "_XmlText", XmlText_t  }\n'
            s += indent + "};\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'string_class_map':
        if target == 'python':
            s += indent + "XML2CLASS = {\n"
            for element in elements:
                if 'properties' in element and 'xmltag' in element['properties'] and element['properties']['xmltag']:
                    s += indent + '    "' + element['properties']['xmltag'] + '": ' + element['class'] + ',\n'
            s += indent + "}\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'annotationtype_layerclass_map':
        if target == 'python':
            s += indent + "ANNOTATIONTYPE2LAYERCLASS = {\n"
            for element in elements:
                if element['class'].endswith('Layer'):
                    if 'properties' in element and 'xmltag' in element['properties'] and element['properties']['xmltag'] and 'annotationtype' in element['properties']:
                        s += indent + "    AnnotationType." + element['properties']['annotationtype'] + ':  ' + element['class'] + ' ,\n'
            s += indent + "    AnnotationType.PREDICATE:  SemanticRolesLayer\n"
            s += indent + "}"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'default_ignore':
        if target == 'c++':
            s += indent + "const set<ElementType> default_ignore = { " + ", ".join([ e + '_t' for e in sorted(flattenclasses(spec['default_ignore'])) ]) + " };\n"
        elif target == 'python':
            s += indent + "default_ignore = ( " + ", ".join(spec['default_ignore']) + ",)\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'default_ignore_annotations':
        if target == 'c++':
            s += indent + "const set<ElementType> default_ignore_annotations = { " + ", ".join([ e + '_t' for e in sorted(flattenclasses(spec['default_ignore_annotations'])) ]) + " };\n"
        elif target == 'python':
            s += indent + "default_ignore_annotations = ( " + ", ".join(spec['default_ignore_annotations']) + ",)\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'default_ignore_structure':
        if target == 'c++':
            s += indent + "const set<ElementType> default_ignore_structure = { " + ", ".join([ e + '_t' for e in sorted(flattenclasses(spec['default_ignore_structure'])) ]) + " };\n"
        elif target == 'python':
            s += indent + "default_ignore_structure = ( " + ", ".join(spec['default_ignore_structure']) + ",)\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'typehierarchy':
        if target == 'c++':
            s += indent + "static const map<ElementType, set<ElementType> > typeHierarchy = { "
            for child, parentset in sorted(parents.items()):
                s += indent + "   { " + child + '_t' + ", { " + ",".join([p + '_t' for p in parentset ]) + " } },\n"
            s += indent + "   { PlaceHolder_t , { Word_t, AbstractStructureElement_t } }\n"
            s += indent + "};\n";
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block == 'attributefeatures':
        if target == 'c++':
            l = []
            for element in elements:
                if 'properties' in element and 'subset' in element['properties'] and element['properties']['subset']:
                    if element['class'] == 'HeadFeature':
                        l.append("headfeature")
                    else:
                        l.append(element['properties']['subset'])
            l.sort()
            s += indent + "const set<string> AttributeFeatures = { " + ", ".join([ '"' + x + '"' for x in l ]) + " };\n"
        else:
            raise NotImplementedError("Block " + block + " not implemented for " + target)
    elif block in spec:
        #simple variable blocks
        s += indent + outputvar(varname, spec[block], target, True)
    else:
        raise Exception("No such block exists in foliaspec: " + block)


    if s and s[-1] != '\n': s += '\n'
    return s


def parser(filename):
    if filename[-2:] in ('.h','.c') or filename[-4:] in ('.cxx','.cpp','.hpp'):
        target = 'c++' #libfolia
        commentsign = '//'
    elif filename[-3:] == '.py':
        target = 'python' #pynlpl.formats.folia
        commentsign = '#'
    else:
        raise Exception("No target language could be deduced from the filename " + filename)

    if not os.path.exists(filename):
        raise FileNotFoundError("File not found: " + filename)

    out = open(filename+'.foliaspec.out','w',encoding='utf-8')


    inblock = False
    blockname = blocktype = ""
    indent = ""
    with open(filename,'r',encoding='utf-8') as f:
        for line in f:
            strippedline = line.strip()
            if not inblock:
                if strippedline.startswith(commentsign + 'foliaspec:'):
                    indent = line.find(strippedline) * ' '
                    fields = strippedline[len(commentsign):].split(':')
                    if fields[1] in ('begin','start'):
                        blocktype = 'explicit'
                        blockname = fields[2]
                        try:
                            varname = fields[3]
                        except:
                            varname = blockname
                    elif len(fields) >= 2:
                        blocktype = 'implicit'
                        blockname = fields[1]
                        try:
                            varname = fields[2]
                        except:
                            varname = blockname
                    else:
                        raise Exception("Syntax error: " + strippedline)
                    inblock = True
                    out.write(line)
                elif strippedline.split(' ')[-1].startswith(commentsign + 'foliaspec:'):
                    fields = strippedline.split(' ')[-1][len(commentsign):].split(':')
                    blocktype = 'line'
                    blockname = fields[1]
                    try:
                        varname = fields[2]
                    except:
                        varname = blockname
                    if varname != blockname:
                        out.write( outputblock(blockname, target, varname) + " " + commentsign + "foliaspec:" + blockname + ":" + varname + "\n")
                    else:
                        out.write( outputblock(blockname, target, varname) + " " + commentsign + "foliaspec:" + blockname + "\n")
                else:
                    out.write(line)
            else:
                if not strippedline and blocktype == 'implicit':
                    out.write(outputblock(blockname, target, varname,indent) + "\n")
                    inblock = False
                elif blocktype == 'explicit' and strippedline.startswith(commentsign + 'foliaspec:end:'):
                    out.write(outputblock(blockname, target, varname,indent) + "\n" + commentsign + "foliaspec:end:" + blockname + "\n")
                    inblock = False

    os.rename(filename+'.foliaspec.out', filename)

def usage():
    print("Syntax: foliaspec.py [filename] [filename] ..etc.." ,file=sys.stderr)
    print("Filenames are Python or C++ files that may contain foliaspec instructions, the files will be updated according to the latest specification in folia.yml",file=sys.stderr)
    sys.exit(0)

def main():
    if len(sys.argv) == 1:
        usage()

    for filename in sys.argv[1:]:
        if filename in ('-h', '--help'):
            usage()
        elif filename in ('-v', '--version'):
            print("FoLiA specification is at version v" + spec['version'],file=sys.stderr)
            sys.exit(0)

        parser(filename)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-


from __future__ import print_function, unicode_literals, division, absolute_import

import getopt
import sys
try:
    from pynlpl.formats import folia
except:
    print("ERROR: pynlpl not found, please obtain PyNLPL from the Python Package Manager ($ sudo easy_install pynlpl) or directly from github: $ git clone git://github.com/proycon/pynlpl.git",file=sys.stderr)
    sys.exit(2)

def usage():
    print("foliacat",file=sys.stderr)
    print("  by Maarten van Gompel (proycon)",file=sys.stderr)
    print("  Radboud University Nijmegen",file=sys.stderr)
    print("  2014 - Licensed under GPLv3",file=sys.stderr)
    print("",file=sys.stderr)
    print("Concatenates multiple FoLiA documents into one; provided that all IDs are unique.",file=sys.stderr)
    print("",file=sys.stderr)
    print("Usage: foliacat [options] file1 file2 file3 ... ",file=sys.stderr)
    print("",file=sys.stderr)
    print("Options:",file=sys.stderr)
    print("  -o [file]                    Output file",file=sys.stderr)
    print("  -i [id]                      ID for output file (mandatory)",file=sys.stderr)


def concat(target, source):
    merges = 0
    for e in source:
        c = e.copy(target.doc)
        target.append(c)
        merges += 1
    return merges



def foliacat(id, outputfile, *files):
    totalmerges = 0
    outputdoc = folia.Document(id=id)
    text = outputdoc.append(folia.Text(outputdoc,id=id + ".text"))
    for i, filename in enumerate(files):
        merges = 0
        print("Processing " + filename, file=sys.stderr)
        inputdoc = folia.Document(file=filename)
        print("(merging document)",file=sys.stderr)

        for annotationtype,set in inputdoc.annotations:
            if not outputdoc.declared(annotationtype,set):
                outputdoc.declare( annotationtype, set)

        for d in inputdoc.data:
            merges += concat(text, d)

        print("(merged " + str(merges) + " elements, with all elements contained therein)",file=sys.stderr)
        totalmerges += merges

    print("(TOTAL: merged " + str(totalmerges) + " elements, with all elements contained therein)",file=sys.stderr)
    if outputfile and merges > 0:
        outputdoc.save(outputfile)

    return outputdoc

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:i:h", ["help"])
    except getopt.GetoptError as err:
        print(str(err),file=sys.stderr)
        usage()
        sys.exit(2)

    outputfile = None
    substitute = False

    id = None
    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '-o':
            outputfile = a
        elif o == '-i':
            id = a
        else:
            raise Exception("No such option: " + o)

    if len(args) < 2:
        print("WARNING: only one file specified", file=sys.stderr)
    if not id:
        print("ERROR: Please specify an ID for the result document with the -i option",file=sys.stderr)
        sys.exit(2)


    if substitute:
        outputfile = args[0]

    outputdoc = foliacat(id, outputfile, *args)
    if not outputfile:
        xml = outputdoc.xmlstring()
        if sys.version < '3':
            if isinstance(xml,unicode):
                print(xml.encode('utf-8'))
            else:
                print(xml)
        else:
            print(xml)


if __name__ == "__main__":
    main()






#! /usr/bin/env python
# -*- coding: utf8 -*-


from __future__ import print_function, unicode_literals, division, absolute_import

import getopt
import io
import sys
import os
import glob
try:
    from pynlpl.formats import folia
except:
    print("ERROR: pynlpl not found, please obtain PyNLPL from the Python Package Manager ($ sudo easy_install pynlpl) or directly from github: $ git clone git://github.com/proycon/pynlpl.git",file=sys.stderr)
    sys.exit(2)

def usage():
    print("foliatextcontent",file=sys.stderr)
    print("  by Maarten van Gompel (proycon)",file=sys.stderr)
    print("  Radboud University Nijmegen",file=sys.stderr)
    print("  2015 - Licensed under GPLv3",file=sys.stderr)
    print("",file=sys.stderr)
    print("This tool operates on some of the redundancy regarding text context inherent in FoLiA documents. It adds text content elements,  on the higher (untokenised) levels, adding offset information and mark-up element if present. Secondly, the tool may als adds text-markup elements for substrings (str element) (provided there is no overlap).",file=sys.stderr)
    print("",file=sys.stderr)
    print("Usage: foliatextcontent [options] file-or-dir1 file-or-dir2 ..etc..",file=sys.stderr)
    print("",file=sys.stderr)
    print("Parameters for output:"        ,file=sys.stderr)
    print("  -s                           Add text content on sentence level",file=sys.stderr)
    print("  -p                           Add text content on paragraph level"    ,file=sys.stderr)
    print("  -d                           Add text content on division level",file=sys.stderr)
    print("  -t                           Add text content on global text level"    ,file=sys.stderr)
    print("  -T                           Add text content for the specified elements (comma separated list of folia xml tags)"    ,file=sys.stderr)
    print("  -X                           Do NOT add offset information"    ,file=sys.stderr)
    print("  -F                           Force offsets to refer to the specified structure only (only works if you specified a single element type for -T!!!)"    ,file=sys.stderr)
    print("  -M                           Add substring markup linking to string elements (if any, and when there is no overlap)"    ,file=sys.stderr)
    print("  -e [encoding]                Output encoding (default: utf-8)",file=sys.stderr)
    print("  -w                           Edit file(s) (overwrites input files), will output to stdout otherwise" ,file=sys.stderr)
    print("Parameters for processing directories:",file=sys.stderr)
    print("  -r                           Process recursively",file=sys.stderr)
    print("  -E [extension]               Set extension (default: xml)",file=sys.stderr)


def linkstrings(element, cls='current',debug=False):
    if element.hastext(cls,strict=True) and element.hasannotation(folia.String):
        text = element.textcontent(cls, correctionhandling=folia.CorrectionHandling.EITHER)

        for string in element.select(folia.String, None, False):
            try:
                stringtextcontent = string.textcontent(cls, correctionhandling=folia.CorrectionHandling.EITHER)
                stringtext = stringtextcontent.text()
                stringoffset = stringtextcontent.offset
            except folia.NoSuchText:
                continue

            if not stringtext:
                continue

            if debug: print("Finding string '" + stringtext + "' in text: ", text.text(), file=sys.stderr)

            offset = 0 #current offset cursor
            length = len(stringtext)
            replaceindex = 0
            replace = []
            for i, subtext in enumerate(text):
                if isinstance(subtext, str):
                    subtextlength = len(subtext)
                    if stringoffset >= offset and stringoffset+length <= offset+subtextlength:
                        reloffset = stringoffset-offset

                        if subtext[reloffset:reloffset+length] != stringtext:
                            print(" String refers to offset " + str(stringoffset) + ", but is not found there ! Found '" + subtext[reloffset:reloffset+length] + "' instead.",file=sys.stderr)
                        else:
                            #match!
                            kwargs = {}
                            replaceindex = i
                            if string.id:
                                kwargs['idref'] = string.id
                            replace = [subtext[:reloffset], folia.TextMarkupString(element.doc, *stringtext, **kwargs), subtext[reloffset+length:]]
                            break

                elif isinstance(subtext, folia.AbstractTextMarkup):
                    subtextlength = len(subtext.text())
                    if stringoffset >= offset and stringoffset+length <= offset+subtextlength:
                        raise NotImplementedError("String fits within other markup element (possibly another string), not implemented yet")

                offset += subtextlength

            if replace:
                if debug: print("Replacing item " + str(replaceindex) + " with: ", replace,file=sys.stderr)
                del text.data[replaceindex]
                for x in reversed(replace):
                    if x:
                        text.data.insert(replaceindex,x)
            else:
                if string.id:
                    print("Could not find string " + string.id + " !!!",file=sys.stderr)

def gettextsequence(element, cls, debug=False):
    assert element.PRINTABLE
    if debug: print(" Getting text for ", repr(element),file=sys.stderr)
    if element.TEXTCONTAINER:
        if debug: print("  Found textcontainer ", repr(element), "in", repr(element.ancestor(folia.AbstractStructureElement)),file=sys.stderr)

        if isinstance(element,folia.TextContent) and element.cls != cls:
            if debug: print("  Class mismatch", element.cls,"vs",cls,file=sys.stderr)
            raise StopIteration

        for e in element:
            if isinstance(e, str):
                if debug: print("  Found: ", e,file=sys.stderr)
                yield e, element
            else: #markup (don't recurse)
                if debug: print("  Found markup: ", repr(e),file=sys.stderr)
                yield e, element
                yield e.gettextdelimiter(), None

        yield None,None #Signals a break after this, if we have text content we needn't delve deeper
    else:
        #Do we have a text content?
        foundtext = False
        if debug: print(" Looking for text in ", repr(element),file=sys.stderr)
        for e in element:
            if isinstance(e, folia.TextContent) and e.cls == cls:
                foundtext = True
                for x in gettextsequence(e, cls, debug):
                    yield x
            elif isinstance(e, folia.Correction):
                foundtextincorrection =False
                try:
                    if e.hasnew() and e.new().textcontent(cls):
                        foundtextincorrection = True
                        for x in gettextsequence(e.new().textcontent(cls), cls, debug):
                            yield x
                except folia.NoSuchText:
                    pass
                except folia.NoSuchAnnotation:
                    pass
                if not foundtextincorrection:
                    try:
                        if e.hascurrent() and e.current().textcontent(cls):
                            foundtextincorrection = True
                            for x in gettextsequence(e.current().textcontent(cls), cls, debug):
                                yield x
                    except folia.NoSuchText:
                        pass
                    except folia.NoSuchAnnotation:
                        pass
                if not foundtextincorrection:
                    try:
                        if e.hasoriginal() and e.original().textcontent(cls):
                            foundtextincorrection = True
                            for x in gettextsequence(e.current().textcontent(cls), cls, debug):
                                yield x
                    except folia.NoSuchText:
                        pass
                    except folia.NoSuchAnnotation:
                        pass
                foundtext = foundtextincorrection

        if not foundtext:
            if debug: print(" Looking for text in children of ", repr(element),file=sys.stderr)
            for e in element:
                if e.PRINTABLE and not isinstance(e, folia.String):
                    #abort = False
                    for x in gettextsequence(e, cls, debug):
                        foundtext = True
                        if x[0] is None:
                            abort = True
                            break
                        yield x
                    #if abort:
                    #    print(" Abort signal received, not processing further elements in ", repr(element),file=sys.stderr)
                    #    break
                if foundtext:
                    delimiter = e.gettextdelimiter()
                    if debug: print(" Got delimiter " + repr(delimiter) + " from " + repr(element), file=sys.stderr)
                    yield e.gettextdelimiter(), None
                elif isinstance(e, folia.AbstractStructureElement) and not isinstance(e, folia.Linebreak) and not isinstance(e, folia.Whitespace):
                    raise folia.NoSuchText("No text was found in the scope of the structure element")


def settext(element, cls='current', offsets=True, forceoffsetref=False, debug=False):
    assert element.PRINTABLE

    if debug: print("In settext for  ", repr(element),file=sys.stderr)

    #get the raw text sequence
    try:
        textsequence = list(gettextsequence(element,cls,debug))
    except folia.NoSuchText:
        return None

    if debug: print("Raw text:  ", textsequence,file=sys.stderr)

    if textsequence:
        newtextsequence = []
        offset = 0
        prevsrc = None
        for i, (e, src) in enumerate(textsequence):
            if e: #filter out empty strings
                if isinstance(e,str):
                    length = len(e)

                    #only whitespace from here on?
                    if not e.strip():
                        onlywhitespace = True
                        for x,y in textsequence[i+1:]:
                            if y is not None:
                                onlywhitespace = False
                        if onlywhitespace:
                            break
                elif isinstance(e, folia.AbstractTextMarkup):
                    e = e.copy()
                    length = len(e.text())

                if src and offsets and src is not prevsrc:
                    ancestors = list(src.ancestors(folia.AbstractStructureElement))
                    if len(ancestors) >= 2 and ancestors[1] is element:
                        if debug: print("Setting offset for text in  " + repr(ancestors[0]) + " to " + str(offset) + ", reference " + repr(element) ,file=sys.stderr)
                        src.offset = offset
                    elif forceoffsetref:
                        src.offset = offset
                        src.ref = element
                    prevsrc = src

                newtextsequence.append(e)
                offset += length

        if newtextsequence:
            if debug: print("Setting text for " + repr(element) + ":" , newtextsequence, file=sys.stderr)
            return element.replace(folia.TextContent, *newtextsequence, cls=cls) #appends if new



def processelement(element, settings):
    if not isinstance(element, folia.AbstractSpanAnnotation): #prevent infinite recursion
        for e in element:
            if isinstance(e, folia.AbstractElement):
                if settings.debug: print("Processing ", repr(e),file=sys.stderr)
                processelement(e,settings)
        if element.PRINTABLE:
            if any( isinstance(element,C) for C in settings.Classes):
                for cls in element.doc.textclasses:
                    settext(element, cls, settings.offsets, settings.forceoffsetref, settings.debug)


def process(filename, outputfile = None):
    print("Converting " + filename,file=sys.stderr)
    doc = folia.Document(file=filename)

    if settings.linkstrings:
        for element in doc.select(folia.AbstractStructureElement):
            if settings.linkstrings:
                for cls in element.doc.textclasses:
                    linkstrings(element, cls, settings.debug)

    if settings.Classes:
        for e in doc.data:
            processelement(e, settings)


    if settings.inplaceedit:
        doc.save()
    else:
        print(doc.xmlstring())

def processdir(d, outputfile = None):
    print("Searching in  " + d, file=sys.stderr)
    for f in glob.glob(os.path.join(d ,'*')):
        if f[-len(settings.extension) - 1:] == '.' + settings.extension:
            process(f, outputfile)
        elif settings.recurse and os.path.isdir(f):
            processdir(f)



class settings:
    Classes = []
    inplaceedit = False
    offsets = True
    forceoffsetref = False
    linkstrings = False

    extension = 'xml'
    recurse = False
    encoding = 'utf-8'

    debug = False

    textclasses =[]


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "E:hsSpPdDtTXMe:wT:Fc:", ["help"])
    except getopt.GetoptError as err:
        print(str(err),file=sys.stderr)
        usage()
        sys.exit(2)


    outputfile = None


    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '-d':
            settings.Classes.append(folia.Division)
        elif o == '-t':
            settings.Classes.append(folia.Text)
        elif o == '-s':
            settings.Classes.append(folia.Sentence)
        elif o == '-p':
            settings.Classes.append(folia.Paragraph)
        elif o == '-T':
            settings.Classes += [ folia.XML2CLASS[tag] for tag in a.split(',') ]
        elif o == '-X':
            settings.offsets = False
        elif o == '-e':
            settings.encoding = a
        elif o == '-E':
            settings.extension = a
        elif o == '-F':
            settings.forceoffsetref = True
        elif o == '-M':
            settings.linkstrings = True
        elif o == '-w':
            settings.inplaceedit = True
        elif o == '-r':
            settings.recurse = True
        elif o == '-D':
            settings.debug = True
        else:
            raise Exception("No such option: " + o)


    if outputfile: outputfile = io.open(outputfile,'w',encoding=settings.encoding)

    if len(settings.Classes) > 1:
        settings.forceoffsetref = False

    if args:
        for x in args:
            if os.path.isdir(x):
                processdir(x,outputfile)
            elif os.path.isfile(x):
                process(x, outputfile)
            else:
                print("ERROR: File or directory not found: " + x,file=sys.stderr)
                sys.exit(3)
    else:
        print("ERROR: Nothing to do, specify one or more files or directories",file=sys.stderr)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

#---------------------------------------------------------------
# ReStructuredText to FoLiA Converter
#   by Maarten van Gompel
#   Centre for Language Studies
#   Radboud University Nijmegen
#   proycon AT anaproy DOT nl
#
#   Licensed under GPLv3
#
# This script converts RST to FoLiA format.
#
#----------------------------------------------------------------

from __future__ import print_function, unicode_literals, division, absolute_import

import sys
import glob
import gzip
import os

from collections import defaultdict
from copy import copy

from docutils import writers, nodes
from docutils.core import publish_cmdline, default_description

try:
    import locale
    locale.setlocale(locale.LC_ALL, '')
except:
    pass

LIBVERSION = "0.1"

class Writer(writers.Writer):

    DEFAULTID = "untitled"
    TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
%(stylesheet)s
<FoLiA xmlns="http://ilk.uvt.nl/folia" xmlns:xlink="http://www.w3.org/1999/xlink" xml:id="%(docid)s" version="0.11.3" generator="docutils-rst2folia-%(libversion)s">
<metadata type="native">
 <annotations>
%(declarations)s
 </annotations>
%(metadata)s
</metadata>
%(content)s
</FoLiA>
"""

    DEFAULTSTYLESHEET = "folia2html.xsl"

    DEFAULTSETS = {
        'division': 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/divisions.foliaset.xml',
        'style': 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/styles.foliaset.xml',
        'note': 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/notes.foliaset.xml',
        'gap': 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/gaps.foliaset.xml',
        'string': None,
    }

    #Formats this writer supports
    supported = ('folia',)

    settings_spec = (
        'FoLiA-Specific Options',
        None,
        (
            ('Document ID.  Default is "%s".' % DEFAULTID, ['--docid'], {'default': DEFAULTID, 'metavar': '<string>'}),
            ('Parent ID. Assign IDs under the specified element, this can be used to merge output back into a larger document', ['--parentid'], {'metavar': '<string>'}),
            ('Parent Type. Assume all new elements start under an element of this type (FoLiA tag), this can be used to merge output back into a larger document, use with --parentid', ['--parenttype'], {'default': 'div', 'metavar': '<string>'}),
            ("Excerpt only. Output only the text node and all elements under it. No standalone document, results may be inserted verbatim into a larger document if used with --parentid/--parenttype and --declare-all", ['--excerpt'], {'default': False, 'action': 'store_true'}),
            ("Declare all possible sets, even if they're not used.", ['--declare-all'], {'default': False, 'action': 'store_true'}),
            ("Sets. Comma separated list of annotationtype:seturl pairs. Example: division:https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/divisions.foliaset.xml", ['--sets'],{'default':""}),
            ("Stylesheet. XSL Stylesheet to associate with the document. Defaults to '%s'" % DEFAULTSTYLESHEET, ['--stylesheet'], {'default': "folia2html.xsl",'metavar':'<string>'}),
        )
    )

    visitor_attributes = ('declarations','metadata','content')

    def translate(self):
        sets = copy(self.DEFAULTSETS)
        for setassignment in self.document.settings.sets.split(','):
            if setassignment:
                annotationtype,set = setassignment.split(':')
                sets[annotationtype] = set
        self.visitor =  FoLiATranslator(self.document, sets)
        self.document.walkabout(self.visitor)
        for attr in self.visitor_attributes:
            setattr(self, attr, getattr(self.visitor, attr))
        self.output = self.apply_template()

    def apply_template(self):
        subs = self.interpolation_dict()
        if self.document.settings.excerpt:
            return "%(content)s" % subs
        else:
            return self.TEMPLATE % subs

    def interpolation_dict(self):
        subs = {}
        for attr in self.visitor_attributes:
            subs[attr] = ''.join(getattr(self, attr)).rstrip('\n')
        subs['encoding'] = self.document.settings.output_encoding
        subs['libversion'] = LIBVERSION
        subs['docid'] = self.document.settings.docid
        subs['stylesheet'] =  "<?xml-stylesheet type=\"text/xsl\" href=\"" + self.document.settings.stylesheet + "\"?>"
        return subs

    def assemble_parts(self):
        writers.Writer.assemble_parts(self)
        for part in self.visitor_attributes:
            self.parts[part] = ''.join(getattr(self, part))



class FoLiATranslator(nodes.NodeVisitor):


    def __init__(self, document, sets={}):
        self.textbuffer = []
        self.path = [] #(tag, id) tuples of the current FoLiA path
        self.content = [] #will contain all XML content as strings
        self.metadata = []
        self.declarations = []
        self.id_store = defaultdict( lambda: defaultdict(int) )
        self.docid = document.settings.docid
        self.list_enumerated = [] #contains a 2-list of boolean, int pairs, indicating whether the list is enumerated or not, and the number of items in it thus-far (used for labels), support nesting.
        self.rootdiv = False #create a root div element?
        self.sets = sets
        self.declared = {}
        self.texthandled = False
        if document.settings.declare_all:
            for key in self.sets:
                self.declare(key)
        if document.settings.parentid:
            self.parentid = document.settings.parentid
            self.path.append( (document.settings.parenttype, self.parentid ) )
            self.textid = "temporary-container-only"
        else:
            self.textid = self.docid + ".text"
            self.parentid = None
        self.excerpt = document.settings.excerpt


        nodes.NodeVisitor.__init__(self, document)

    ############# HELPERS ###############

    def astext(self):
        return ''.join(self.head + self.content)

    def encode(self, text):
        """Encode special characters in `text` & return."""
        if sys.version < '3' and not isinstance(text, unicode):
            text = unicode(text, 'utf-8')
        elif sys.version >= '3' and not isinstance(text, str):
            text = str(text, 'utf-8')
        return text.translate({
            ord('&'): '&amp;',
            ord('<'): '&lt;',
            ord('>'): '&gt;',
        })

    def initstructure(self, tag, **attribs):
        """Generic visit function for structure elements"""
        #Generate an ID
        if tag == "text":
            id = self.textid
        else:
            parenttag, parentid = self.path[-1]
            id = self.generate_id(parentid, tag)
        self.declare(tag)
        self.path.append( (tag, id ) )
        indentation = (len(self.path)-1) * " "
        o = indentation + "<" + tag + " xml:id=\"" + id + "\""
        if tag == "text" and self.excerpt: #this is the root of our output, add namespace stuff
            o += ' xmlns="http://ilk.uvt.nl/folia" xmlns:xlink="http://www.w3.org/1999/xlink"'
        if attribs:
            for key, value in attribs.items():
                if key == "cls": key = "class"
                if sys.version < '3':
                    o += " " + key + "=\"" + unicode(value) + "\""
                elif sys.version >= '3':
                    o += " " + key + "=\"" + str(value) + "\""
        o += ">\n"
        self.content.append(o)

    def closestructure(self, tag):
        """Generic depart function for structure elements"""
        _tag, id = self.path.pop()
        if not tag == _tag:
            raise Exception("Mismatch in closestructure, expected closure for " + tag + ", got " + _tag)
        indentation = len(self.path) * " "
        o = ""
        if self.textbuffer:
            o += indentation + " <t>"  + "".join(self.textbuffer) + "</t>\n"
        o += indentation + "</" + tag + ">\n"
        self.textbuffer = []
        self.content.append(o)

    def generate_id(self, parentid, tag ):
        if parentid == "temporary-container-only" and self.parentid:
            self.id_store[self.parentid][tag] += 1
            return self.parentid + "." + tag + "." + str(self.id_store[parentid][tag])
        else:
            self.id_store[parentid][tag] += 1
            return parentid + "." + tag + "." + str(self.id_store[parentid][tag])


    def rightsibling(self, node):
        fetch = False
        for sibling in node.traverse(None,1,0,1,0):
            if sibling is node:
                fetch = True
            elif fetch:
                return sibling
        return None


    def ignore_depart(self, node):
        try:
            if node.ignore_depart:
                return True
        except AttributeError:
            return False

    def addstyle(self,node,style):
        self.texthandled = True
        self.declare('style')
        self.textbuffer.append(  '<t-style class="' + style + '">' + self.encode(node.astext()) + '</t-style>' )

    def addlink(self,node,url):
        self.texthandled = True
        self.declare('string')
        self.textbuffer.append(  '<t-str xlink:type="simple" xlink:href="' + url + '">' + self.encode(node.astext()) + '</t-str>' )

    def addmetadata(self, key, node):
        self.texthandled = True
        self.metadata.append(  " <meta id=\"" + key + "\">" + self.encode(node.astext()) + "</meta>\n" )


    def declare(self, annotationtype):
        if annotationtype == 'div':
            annotationtype = 'division'
        elif annotationtype == 's':
            annotationtype = 'sentence'
        elif annotationtype == 'p':
            annotationtype = 'paragraph'
        if not annotationtype in self.declared:
            if annotationtype in self.sets:
                if self.sets[annotationtype]:
                    self.declarations.append("   <" + annotationtype + "-annotation set=\"" + self.sets[annotationtype] + "\" />\n")
                else:
                    self.declarations.append("   <" + annotationtype + "-annotation />\n")
                self.declared[annotationtype] = True

    ############# TRANSLATION HOOKS (MAIN STRUCTURE) ################


    def visit_document(self, node):
        self.initstructure('text')

    def depart_document(self, node):
        if self.rootdiv:
            self.closestructure('div')
        self.closestructure('text')

    def visit_paragraph(self, node):
        if node.parent.__class__.__name__ == 'list_item':
            #this paragraph is in an item, we don't want paragraphs in items unless there actually are multiple elements in the item
            sibling = self.rightsibling(node)
            if sibling:
                self.initstructure('p')
            else:
                node.ignore_depart = True
        else:
            self.initstructure('p')

    def depart_paragraph(self, node):
        if not self.ignore_depart(node):
            self.closestructure('p')

    def visit_section(self, node):
        self.initstructure('div',cls="section")

    def depart_section(self, node):
        self.closestructure('div')

    def visit_title(self, node):
        if node.parent.__class__.__name__ == 'document':
            self.rootdiv = True
            self.initstructure('div',cls="document" if not self.parentid else "section")
        self.initstructure('head')

    def depart_title(self, node):
        self.closestructure('head')

    def visit_bullet_list(self,node):
        self.list_enumerated.append([False,0])
        self.initstructure('list')

    def depart_bullet_list(self,node):
        self.list_enumerated.pop()
        self.closestructure('list')

    def visit_enumerated_list(self,node):
        self.list_enumerated.append([True,0])
        self.initstructure('list')

    def depart_enumerated_list(self,node):
        self.list_enumerated.pop()
        self.closestructure('list')

    def visit_list_item(self,node):
        if self.list_enumerated[-1][0]:
            self.list_enumerated[-1][1] += 1
            self.initstructure('item',n=self.list_enumerated[-1][1])
        else:
            self.initstructure('item')

    def depart_list_item(self,node):
        self.closestructure('item')

    def visit_image(self,node):
        self.initstructure('figure',src=node['uri'])
    def depart_image(self,node):
        #parent figure will do the closing if image in figure
        if node.parent.__class__.__name__ != "figure":
            self.closestructure('figure')

    def visit_figure(self,node):
        pass
    def depart_figure(self,node):
        self.closestructure('figure')

    def visit_caption(self,node):
        self.initstructure('caption')
    def depart_caption(self,node):
        self.closestructure('caption')


    def visit_literal_block(self,node):
        self.initstructure('gap',cls="verbatim")
        self.texthandled = True
    def depart_literal_block(self,node):
        tag = "gap"
        _tag, id = self.path.pop()
        if not tag == _tag:
            raise Exception("Mismatch in closestructure, expected closure for " + tag + ", got " + _tag)
        indentation = len(self.path) * " "
        o = indentation + " <content><![CDATA["  + node.astext() + "]]></content>\n"
        o += indentation + "</" + tag + ">\n"
        self.content.append(o)
        self.texthandled = False

    def visit_block_quote(self, node):
        self.initstructure('quote')
    def depart_block_quote(self, node):
        self.closestructure('quote')

    ############# TRANSLATION HOOKS (TEXT & MARKUP) ################

    def visit_Text(self, node):
        if not self.texthandled:
            self.textbuffer.append(  self.encode(node.astext()) )

    def depart_Text(self, node):
        pass

    def visit_strong(self, node):
        self.addstyle(node,"strong")
    def depart_strong(self, node):
        self.texthandled = False

    def visit_emphasis(self, node):
        self.addstyle(node,"emphasis")
    def depart_emphasis(self, node):
        self.texthandled = False

    def visit_literal(self, node):
        self.addstyle(node,"literal")
    def depart_literal(self, node):
        self.texthandled = False

    def visit_reference(self, node):
        self.addlink(node,node.attributes['refuri'])
    def depart_reference(self, node):
        self.texthandled = False

    def visit_target(self, node): #TODO? Seems to work, am I missing something?
        pass
    def depart_target(self, node):
        pass

    def visit_comment(self, node):
        self.texthandled = True
    def depart_comment(self, node):
        self.content.append("<!-- " + node.astext() + " -->\n")
        self.texthandled = False


    ############# TRANSLATION HOOKS (OTHER STRUCTURE) ################

    def visit_footnote(self,node):
        #TODO: handle footnote numbering:  http://code.nabla.net/doc/docutils/api/docutils/transforms/references/docutils.transforms.references.Footnotes.html
        self.initstructure('note',cls='footnote')
    def depart_footnote(self,node):
        self.closestructure('note')

    def visit_attention(self,node):
        self.initstructure('note',cls='attention')
    def depart_attention(self,node):
        self.initstructure('note')

    def visit_hint(self,node):
        self.initstructure('note',cls='hint')
    def depart_hint(self,node):
        self.closestructure('note')


    def visit_note(self,node):
        self.initstructure('note',cls='note')
    def depart_note(self,node):
        self.closestructure('note')

    def visit_caution(self,node):
        self.initstructure('note',cls='caution')
    def depart_caution(self,node):
        self.closestructure('note')

    def visit_warning(self,node):
        self.initstructure('note',cls='warning')
    def depart_warning(self,node):
        self.closestructure('note')

    def visit_danger(self,node):
        self.initstructure('note',cls='danger')
    def depart_danger(self,node):
        self.closestructure('note')

    def visit_admonition(self,node):
        self.initstructure('note',cls='admonition')
    def depart_admonition(self,node):
        self.closestructure('note')

    def visit_tip(self,node):
        self.initstructure('note',cls='tip')
    def depart_tip(self,node):
        self.closestructure('note')

    def visit_error(self,node):
        self.initstructure('note',cls='error')
    def depart_error(self,node):
        self.closestructure('note')

    def visit_important(self,node):
        self.initstructure('note',cls='important')
    def depart_important(self,node):
        self.closestructure('note')
    ############# TRANSLATION HOOKS (METADATA, rst-specific fields) ################

    def visit_docinfo(self, node):
        pass
    def depart_docinfo(self, node):
        pass
    def visit_authors(self, node):
        pass
    def depart_authors(self, node):
        pass

    def visit_author(self, node):
        self.addmetadata('author', node)
    def depart_author(self, node):
        self.texthandled = False

    def visit_date(self, node):
        self.addmetadata('date', node)

    def depart_date(self, node):
        self.texthandled = False

    def visit_contact(self, node):
        self.addmetadata('contact', node)

    def depart_contact(self, node):
        self.texthandled = False

    def visit_status(self, node):
        self.addmetadata('status', node)

    def depart_status(self, node):
        self.texthandled = False


    def visit_version(self, node):
        self.addmetadata('version', node)

    def depart_version(self, node):
        self.texthandled = False

    def visit_copyright(self, node):
        self.addmetadata('copyright', node)

    def depart_copyright(self, node):
        self.texthandled = False


    def visit_organization(self, node):
        self.addmetadata('organization', node)

    def depart_organization(self, node):
        self.texthandled = False



    def visit_address(self, node):
        self.addmetadata('address', node)

    def depart_address(self, node):
        self.texthandled = False


    def visit_contact(self, node):
        self.addmetadata('contact', node)

    def depart_contact(self, node):
        self.texthandled = False


def main():
    description = 'Generates FoLiA documents from reStructuredText. ' + default_description
    publish_cmdline(writer=Writer(), writer_name='folia', description=description)

if __name__ == '__main__':
    main()
